/* A simple DSP filter for the Arduino.
 * Written by Alexis Kotlowy (256byteram@gmail.com)
 * 7-Feb-2013
 *
 * Public Domain
 *
 * This probably has specific timer functions for the
 * ATmega168.
 */


#include <stdint.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "notes.h"

#define SPEAKERPIN 11
#define LEDPIN 13
#define SAMPLE_RATE 16384
#define BPS 4
volatile int soundDat;
volatile int flag;

/* This table has two adjacent entries, one
 * for the note defined in notes.h and the
 * number of beats the note lasts for.
 */
const int music[] PROGMEM = {
  nG4,  1,
  nGb4, 1,
  nE4,  1,
  nE4,  1,
  nGb4, 4,
  -1,   4,
  -1,   3,
  nA3,  1,
  nG4,  1,
  nGb4, 1,
  nE4,  1,
  nE4,  2,
  nGb4, 3,
  nD4,  2,
  nE4,  1,
  nA3,  1,
  -1,   7,
  nA3,  1,
  nE4,  2,
  nGb4, 1,
  nG4,  3,
  nE4,  1,
  nDb4, 2,
  nD4,  3,
  nE4,  2,
  nA3,  1,
  nA3,  2,
  nGb4, 7,
  -1,   4,
  nG4,  1,
  nGb4, 1,
  nE4,  1,
  nE4,  1,
  nGb4, 4,
  -1,   4,
  -1,   3,
  nA3,  1,
  nG4,  1,
  nGb4, 1,
  nE4,  1,
  nE4,  3,
  nGb4, 1,
  nD4,  3,
  nE4,  1,
  nA3,  3,
  -1,   6,
  nE4,  2,  
  nGb4, 1,
  nG4,  3,
  nE4,  1,
  nDb4, 3,
  nD4,  1,
  nE4,  2,
  nA3,  1,
  nD4,  1,
  nE4,  1,
  nF4,  1,
  nE4,  1,
  nD4,  1,
  nC4,  1,
  -1,   4,
  0,0
};

ISR(TIMER1_COMPA_vect) {
  OCR2A = soundDat;
  flag = 1;
}
/* This takes four values:
 * f - the frequency to play at (assuming 16384Hz sample rate)
 * cut - the start cutoff frequency, between 0 and 512.
 * res - the resonance level, between 0 and 256.
 * len - The length of the note in samples.
 */
void makeSound(int f, int cut, int res, uint32_t len) {
  int s = 0;              // Sawtooth waveform counter
  int c, cc;              // Decay counters for filter
  int32_t n1=0, n2=0, i;  // Nodes for filter and length counter
  
  flag = 0;               // Set sample set flag
  TIMSK1 |= _BV(OCIE1A);  // Enable interrupts on timer

  /* This is a poor attempt at scaling the filter cutoff
   * frequency. Essentially, the value needs to move a tiny
   * amount at the low end and a lot at the high end. If you
   * can do a better job, please contact me!
   */
  c = ((int32_t)cut*(int32_t)cut)>>9; 
  
  // Start counting down the samples
  for(i=0;i<len;i++) {

    // The LED is lit if there is a sample underrun
    if (flag != 0) digitalWrite(LEDPIN, HIGH);
    while(flag == 0);     // Wait until sample is updated
    flag = 0;

    s += f<<2;            // Scale from 16 bits to 14 bits (16384)

    // The business end of the program, performs
    // a low pass filter function in integer maths.
    n1 += (c * ((s>>8) - (n1<<1) + ((res * ((n1<<1)- n2))>>8)))>>10;
    n2 += (c * ((n1<<3) - n2))>>10;

    // If a note is specified, update to new sample, else
    // silence the output.
    if (f != -1) soundDat = (n2>>2)+128;
    else soundDat = 128;

    // Count from 0 to 0x7F and when 0 is reached again,
    // decrement the filter cutoff by one, making sure it
    // doesn't go below zero.
    cc = (cc+1) & 0x7F;
    if (cc == 0) c--;
    if (c < 1) c = 1;
  }
  // Reset the output to the middle (unsigned)
  soundDat = 128;
  // Wait for it to be updated
  flag = 0;
  while(flag == 0);
  // Disable timer interrupt.
  TIMSK1 &= ~_BV(OCIE1A);
}

void setup() {
  int i;  
  pinMode(SPEAKERPIN, OUTPUT);
  pinMode(LEDPIN, OUTPUT);

  /* This initialization code is pinched from the
   * Arduino PCMAudio sketch. See 
   * http://playground.arduino.cc/Code/PCMAudio
   */
  ASSR &= ~(_BV(EXCLK) | _BV(AS2));
  TCCR2A = (TCCR2A | _BV(COM2A1)) & ~_BV(COM2A0);
  TCCR2A &= ~(_BV(COM2B1) | _BV(COM2B0));
  TCCR2A |= _BV(WGM21) | _BV(WGM20);
  TCCR2B &= ~_BV(WGM22);
  TCCR2B = (TCCR2B & ~(_BV(CS12) | _BV(CS11))) | _BV(CS10);

  /* Set the PWM output to be in the middle (unsigned)
  OCR2A = 128;
  soundDat = 128;
  /* Initialize flag */
  flag = 0;

  cli();
  /* Set the timer modes, again see the PCMAudio sketch  */
  TCCR1B = (TCCR1B & ~_BV(WGM13)) | _BV(WGM12);
  TCCR1A = TCCR1A & ~(_BV(WGM11) | _BV(WGM10));
  TCCR1B = (TCCR1B & ~(_BV(CS12) | _BV(CS11))) | _BV(CS10);
  OCR1A = F_CPU / SAMPLE_RATE;  
  sei();

}

void loop () {
  int i, n, l;
  
  // Loop for every note in the music table
  for (i=0;;i+=2) {
    // Read note and length
    n = pgm_read_word(&music[i]);
    l = pgm_read_word(&music[i+1]);
    // End when both are zero
    if ((n|l)==0) break;
    // Send it to the sound routine.
    makeSound(n, 300, 250, l*(SAMPLE_RATE/BPS));
  }
  for(;;);  /* Restart the Arduino to get the song to
             * play again. This avoids irritation.
             */
}

